@@ -58,7 +58,7 @@ public:
   Complex&         operator /= (const Complex& y); 
   Complex&         operator /= (double y); 
 
-  void             error(char* msg) const;
+  void             error(const char* msg) const;
 };
 
 // inline members
@@ -113,7 +113,7 @@ inline Complex&  Complex::operator *= (double y)
 
 // error handlers
 
-extern  void default_Complex_error_handler(char*);
+extern  void default_Complex_error_handler(const char*);
 extern  one_arg_error_handler_t Complex_error_handler;
 
 extern  one_arg_error_handler_t 
