@@ -203,7 +203,7 @@ public:
 // error detection
 
   int             initialized() const;
-  volatile void   error(char* msg) const;
+  volatile void   error(const char* msg) const;
   int             OK() const;  
 };
 
