@@ -35,7 +35,7 @@ typedef long int	    ptrdiff_t;
 #ifdef DGUX
 typedef	int	wchar_t;
 #else
-typedef	unsigned char	wchar_t;
+typedef	unsigned short	wchar_t;
 #endif
 
 #ifndef NULL
