@@ -89,7 +89,7 @@ Key_List::get_special_input (char delimiter)
                 }
             }
           else
-            ungetc (c, stdin);
+            buf[i++] = '%';
         }
       else if (i >= size) /* Yikes, time to grow the buffer! */
         { 
@@ -272,7 +272,7 @@ Key_List::read_keys (void)
                       ? "Some input keys have identical hash values, examine output carefully...\n"
                       : "Some input keys have identical hash values,\ntry different key positions or use option -D.\n%a");
       if (option[ALLCHARS])
-        option.set_charset_size (max_key_len);
+        option.set_keysig_size (max_key_len);
     }
 }
 
@@ -421,7 +421,7 @@ Key_List::reorder (void)
 void
 Key_List::output_min_max ()
 {
-  List_Node   *temp;
+  List_Node *temp;
 
   for (temp = head; temp->next; temp = temp->next)
     ;
@@ -429,10 +429,12 @@ Key_List::output_min_max ()
   min_hash_value = head->hash_value;
   max_hash_value = temp->hash_value;
 
-  printf ("\n#define TOTAL_KEYWORDS %d\n#define MIN_WORD_LENGTH %d"
-          "\n#define MAX_WORD_LENGTH %d\n#define MIN_HASH_VALUE %d"
-          "\n#define MAX_HASH_VALUE %d\n/* maximum key range = %d, duplicates = %d */\n\n",
-          total_keys, min_key_len, max_key_len, min_hash_value, max_hash_value,
+  if (option[C])
+    printf ("\n#define TOTAL_KEYWORDS %d\n#define MIN_WORD_LENGTH %d"
+            "\n#define MAX_WORD_LENGTH %d\n#define MIN_HASH_VALUE %d"
+            "\n#define MAX_HASH_VALUE %d\n", total_keys, min_key_len,
+            max_key_len, min_hash_value, max_hash_value);
+  printf ("/* maximum key range = %d, duplicates = %d */\n\n",
           max_hash_value - min_hash_value + 1, total_duplicates);
 }
 
@@ -702,9 +704,15 @@ Key_List::output_hash_function (void)
   if (option[GNU])
     printf ("#ifdef __GNUC__\ninline\n#endif\n");
   
+  if (option[C])
+    printf ("static ");
+  printf ("unsigned int\n");
+  if (option[CPLUSPLUS])
+    printf ("%s::", option.get_class_name ());
+
   printf (option[ANSI] 
-          ? "static int\n%s (register const char *str, register int len)\n{\n  static %sunsigned %s asso_value[] =\n    {"
-          : "static int\n%s (str, len)\n     register char *str;\n     register int unsigned len;\n{\n  static %sunsigned %s asso_value[] =\n    {",
+          ? "%s (register const char *str, register int len)\n{\n  static %sunsigned %s asso_values[] =\n    {"
+          : "%s (str, len)\n     register char *str;\n     register int unsigned len;\n{\n  static %sunsigned %s asso_values[] =\n    {",
           option.get_hash_name (), option[CONST] ? "const " : "",
           max_hash_value <= MAX_UNSIGNED_CHAR ? "char" : (max_hash_value <= MAX_UNSIGNED_SHORT ? "short" : "int"));
   
@@ -718,7 +726,7 @@ Key_List::output_hash_function (void)
   
   /* Optimize special case of ``-k 1,$'' */
   if (option[DEFAULTCHARS]) 
-    printf ("\n    };\n  return %sasso_value[str[len - 1]] + asso_value[str[0]];\n}\n\n",
+    printf ("\n    };\n  return %sasso_values[str[len - 1]] + asso_values[str[0]];\n}\n\n",
             option[NOLENGTH] ? "" : "len + ");
   else
     {
@@ -736,14 +744,14 @@ Key_List::output_hash_function (void)
 
           for (; key_pos != WORD_END; )
             {
-              printf ("asso_value[str[%d]]", key_pos - 1);
+              printf ("asso_values[str[%d]]", key_pos - 1);
               if ((key_pos = option.get ()) != EOS)
                 printf (" + ");
               else
                 break;
             }
 
-          printf ("%s;\n}\n\n", key_pos == WORD_END ? "asso_value[str[len - 1]]" : "");
+          printf ("%s;\n}\n\n", key_pos == WORD_END ? "asso_values[str[len - 1]]" : "");
         }
 
       /* We've got to use the correct, but brute force, technique. */
@@ -758,7 +766,7 @@ Key_List::output_hash_function (void)
               int i;
 
               for (i = max_key_len; i > 0; i--)
-                printf ("      case %d:\n        hval += asso_value[str[%d]];\n", i, i - 1);
+                printf ("      case %d:\n        hval += asso_values[str[%d]];\n", i, i - 1);
               
               printf ("    }\n  return hval;\n}\n\n");
             }
@@ -772,13 +780,13 @@ Key_List::output_hash_function (void)
                   while (--count > key_pos)
                     printf ("      case %d:\n", count);
                   
-                  printf ("      case %d:\n        hval += asso_value[str[%d]];\n",
+                  printf ("      case %d:\n        hval += asso_values[str[%d]];\n",
                           key_pos, key_pos - 1);
                 }
               while ((key_pos = option.get ()) != EOS && key_pos != WORD_END);
               
               printf ("    }\n  return hval%s;\n}\n\n",
-                      key_pos == WORD_END ? " + asso_value[str[len - 1]]" : "");
+                      key_pos == WORD_END ? " + asso_values[str[len - 1]]" : "");
             }
         }
     }
@@ -967,7 +975,7 @@ Key_List::output_lookup_function (void)
 }
 
 /* Generates the hash function and the key word recognizer function
-based upon the user's Options. */
+   based upon the user's Options. */
 
 void 
 Key_List::output (void)
@@ -979,6 +987,15 @@ Key_List::output (void)
     printf ("%s;\n", array_type);
   
   output_min_max ();
+
+  if (option[CPLUSPLUS])
+    printf ("class %s\n{\nprivate:\n  const int MIN_WORD_LENGTH, MAX_WORD_LENGTH, MIN_HASH_VALUE, MAX_HASH_VALUE;\n"
+            "  unsigned int hash (const char *str, int len);\npublic:\n  %s (void):\n"
+            "    MIN_WORD_LENGTH (%d), MAX_WORD_LENGTH (%d), MIN_HASH_VALUE (%d), MAX_HASH_VALUE (%d) {}\n"
+            "  %soperator () (const char *str, int len);\n};\n\n",
+            option.get_class_name (), option.get_class_name (),
+            min_key_len, max_key_len, min_hash_value, max_hash_value, return_type);
+
   output_hash_function ();
   
   if (global_table)
@@ -999,10 +1016,14 @@ Key_List::output (void)
   if (option[GNU])          /* Use the inline keyword to remove function overhead. */
     printf ("#ifdef __GNUC__\ninline\n#endif\n");
   
+  printf ("%s%s\n", option[CONST] ? "const " : "", return_type);
+  if (option[CPLUSPLUS])
+    printf ("%s::", option.get_class_name ());
+
   printf (option[ANSI] 
-          ? "%s%s\n%s (register const char *str, register int len)\n{\n"
-          : "%s%s\n%s (str, len)\n     register char *str;\n     register unsigned int len;\n{\n",
-          option[CONST] ? "const " : "", return_type, option.get_function_name ());
+          ? "%s (register const char *str, register int len)\n{\n"
+          : "%s (str, len)\n     register char *str;\n     register unsigned int len;\n{\n",
+          option[CPLUSPLUS] ? "operator ()" : option.get_function_name ());
   
   /* Use the switch in place of lookup table. */
   if (option[SWITCH])
@@ -1031,12 +1052,9 @@ Key_List::output (void)
     }
   
   if (additional_code)
-    {
-      int c;
-      
-      while ((c = getchar ()) != EOF)
-        putchar (c);
-    }
+    for (int c; (c = getchar ()) != EOF; putchar (c))
+      ;
+
   fflush (stdout);
 }
 
@@ -1056,7 +1074,7 @@ Key_List::sort (void)
 void 
 Key_List::dump () 
 {      
-  int field_width = option.get_charset_size ();
+  int field_width = option.get_max_keysig_size ();
   
   fprintf (stderr, "\nList contents are:\n(hash value, key length, index, %*s, keyword):\n",
            field_width, "char_set");
