@@ -50,6 +50,8 @@ and this notice must be preserved on all copies.
 #define I_MINNUM        ((unsigned long)(I_RADIX >> 1))
 #define I_POSITIVE      1
 #define I_NEGATIVE      0
+
+/* All routines assume SHORT_PER_LONG > 1 */
 #define SHORT_PER_LONG  ((LONGBITS + SHORTBITS - 1) / SHORTBITS)
 #define CHAR_PER_LONG   ((LONGBITS + CHARBITS - 1) / CHARBITS)
 
@@ -343,7 +345,6 @@ IntRep* Icopy_one(IntRep* old, int newsgn)
   return rep;
 }
 
-
 // convert to a legal two's complement long if possible
 // if too big, return most negative/positive value
 
@@ -351,7 +352,19 @@ long Itolong(const IntRep* rep)
 { 
   if (rep->len > SHORT_PER_LONG)
     return (rep->sgn == I_POSITIVE) ? MAXLONG : MINLONG;
-  else
+  else if (rep->len == 0)
+    return 0;
+  else if (rep->len < SHORT_PER_LONG)
+  {
+    unsigned long a = rep->s[rep->len-1];
+    if (SHORT_PER_LONG > 2) // normally optimized out
+    {
+      for (int i = rep->len - 2; i >= 0; --i)
+        a = up(a) | rep->s[i];
+    }
+    return (rep->sgn == I_POSITIVE)? a : -((long)a);
+  }
+  else 
   {
     unsigned long a = rep->s[SHORT_PER_LONG - 1];
     if (a >= I_MINNUM)
@@ -359,7 +372,7 @@ long Itolong(const IntRep* rep)
     else
     {
       a = up(a) | rep->s[SHORT_PER_LONG - 2];
-      if (SHORT_PER_LONG - 3 >= 0)
+      if (SHORT_PER_LONG > 2)
       {
         for (int i = SHORT_PER_LONG - 3; i >= 0; --i)
           a = up(a) | rep->s[i];
@@ -369,7 +382,6 @@ long Itolong(const IntRep* rep)
   }
 }
 
-
 // test whether op long() will work.
 // careful about asymmetry between MINLONG & MAXLONG
 
@@ -2233,7 +2245,7 @@ int Integer::OK() const
   return v;
 }
 
-volatile void Integer::error(char* msg) const
+volatile void Integer::error(const char* msg) const
 {
   (*lib_error_handler)("Integer", msg);
 }
