@@ -34,7 +34,7 @@ and this notice must be preserved on all copies.
 #include <regex.h>
 // }
 
-volatile void String::error(char* msg) const
+volatile void String::error(const char* msg) const
 {
   (void)((*lib_error_handler)("String", msg));
 }
@@ -627,6 +627,7 @@ int String::_gsub(const char* pat, int pl, const char* r, int rl)
       int mustfit = xi + remaining + rl - pl;
       if (mustfit >= nsz)
       {
+        if (nrep != 0) nrep->len = xi;
         nrep = Sresize(nrep, mustfit);
         nsz = nrep->sz;
         x = nrep->s;
@@ -697,6 +698,7 @@ int String::_gsub(const Regex& pat, const char* r, int rl)
       int mustfit = xi + remaining + rl - pl;
       if (mustfit >= nsz)
       {
+        if (nrep != 0) nrep->len = xi;
         nrep = Sresize(nrep, mustfit);
         x = nrep->s;
         nsz = nrep->sz;
@@ -742,7 +744,7 @@ int String::_gsub(const Regex& pat, const char* r, int rl)
 
 void String::del(int pos, int len)
 {
-  if (pos <= 0 || len <= 0 || pos + len > length()) return;
+  if (pos < 0 || len <= 0 || pos + len > length()) return;
   int nlen = length() - len;
   int first = pos + len;
   ncopy0(&(rep->s[first]), &(rep->s[pos]), length() - first);
