@@ -33,8 +33,28 @@ and this notice must be preserved on all copies.
 #define UPDATE_STATS(ARGS)
 #endif
 
+/* History
 
+   Tue Jan 16 04:54:27 1990  Doug Lea  (dl at g.oswego.edu)
 
+     version 1 released in libg++
+
+   Sun Jan 21 05:52:47 1990  Doug Lea  (dl at g.oswego.edu)
+
+     bins are now own struct for, sanity.
+
+     new victim search strategy: scan up and consolidate.
+     Both faster and less fragmentation.
+
+     refined when to scan bins for consolidation, via consollink, etc.
+
+     realloc: always try to expand chunk, avoiding some fragmentation.
+
+     changed a few inlines into macros
+
+     hardwired SBRK_UNIT to 4096 for uniformity across systems
+
+*/
 
 /* 
   A version of malloc/free/realloc tuned for C++ applications.
@@ -64,10 +84,6 @@ and this notice must be preserved on all copies.
  
   This malloc, like any other, is a compromised design. 
 
-  The basic overhead for any chunk of memory is represented in
-  the malloc_chunk struct, which contains 4 fields,
-  although only two of them are used when the chunk is in use
-  (malloced but not yet freed). 
 
   Chunks of memory are maintained using a `boundary tag' method as
   described in e.g., Knuth or Standish.  This means that the size of
@@ -76,6 +92,10 @@ and this notice must be preserved on all copies.
   The size field is also used to hold bits representing whether a
   chunk is free or in use.
 
+  Malloced chunks have space overhead of 8 bytes: The preceding
+  and trailing size fields. When they are freed, the list pointer
+  fields are also needed.
+
   Available chunks are kept in doubly linked lists. The lists are
   maintained in an array of bins using a power-of-two method, except
   that instead of 32 bins (one for each 1 << i), there are 128: each
@@ -90,31 +110,21 @@ and this notice must be preserved on all copies.
   no longer useful, fragments are gathered into bigger chunks awaiting new
   roles.
 
-  The bins av[i] serve as heads of the lists. They are malloc_chunk
-  structs, but  the field `dirty', used only for
-  alignment purposes in chunks, is used to record the number of chunks
-  placed on the list since last consolidation.
+  The bins av[i] serve as heads of the lists. Bins contain a dummy
+  header for the chunk lists, and a `dirty' field used to indicate
+  whether the list may need to be scanned for consolidation.
 
   On allocation, the bin corresponding to the request size is
   scanned, and if there is a chunk with size >= requested, it
-  is used, without splitting off the remainder into a different bin.
-  This helps minimize the useless creation of too many very small unusable 
-  fragments as often occurs with `best fit' style methods. If
-  the bin is not empty, but no candidate chunk is available to
-  fill the request, then the bin is sometimes (according to
-  a heuristic formula) sent for consolidation, in order to
-  try to minimize future useless scans.
+  is split, if too big, and used.
 
-  If no chunk exists, bigger bins are scanned. To minimize 
+  If no chunk exists in the list, a bigger chunk is found. To minimize 
   unsuccessful scanning of a large number of probably empty bins
   (as happens when so many bins are used), a single designated
-  ``victim'' is maintained, and is considered first. The designated
-  victim is generally the last remaindered or consolidated chunk.
+  ``victim'' ibin s maintained, and is considered first. The designated
+  victim is that holding the last remaindered or consolidated chunk.
   If the victim is unusable (too small), then a new victim is
-  searched for by scanning bigger bins. New victims are scanned approximately
-  biggest-first. While this can waste (fragment) big blocks, it is 
-  generally preferable, since the same block is often split/reused
-  many times without requiring searches for new victims.
+  searched for by scanning bigger bins. 
 
   If the victim is usable, it is split. The remainder is placed on
   the back of the appropriate bin list. (All freed chunks are placed
@@ -185,8 +195,15 @@ and this notice must be preserved on all copies.
 
 /* preliminaries */
 
+#ifdef __cplusplus
+#include <stdio.h>
+#else
 #include "//usr/include/stdio.h"  /* needed for error reporting */
-#include <getpagesize.h>          /* needed to find good sbrk size */
+#endif
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 #ifdef USG
 extern void*     memset(void*, int, int);
@@ -199,55 +216,77 @@ extern void      bzero(void*, unsigned int);
 #endif
 
 extern void*     sbrk(unsigned int);
+
+#ifndef __cplusplus
 extern int       fputs(const char*, FILE*);
 extern int       fprintf(FILE*, const char*, ...);
+#endif
 
+#ifdef __GNUC__
 extern volatile void abort();
+#else
+extern void abort();
+#endif
+
+#ifdef __cplusplus
+};  /* end of extern "C" */
+#endif
+
+
+/* A good multiple to call sbrk with */
+
+#define SBRK_UNIT 4096 
+
 
 
 /* how to die on detected error */
+
+#ifdef __GNUC__
 static volatile void malloc_user_error()
+#else
+static void malloc_user_error()
+#endif
 {
   fputs("malloc/free/realloc: clobbered space detected\n", stderr); abort();
 }
 
 
 
-/*
- Basic overhead for each malloc'ed chunk 
-*/
+/*  Basic overhead for each malloc'ed chunk */
 
-/* malloc chunks and malloc bins have the same structure */
 
 struct malloc_chunk
 {
   unsigned int         size;     /* Size in bytes, including overhead. */
                                  /* Or'ed with INUSE if in use. */
-                                 /* For bins, size must == 0  */
 
   struct malloc_chunk* fd;       /* double links -- used only if free. */
-  struct malloc_chunk* bk;       /* For bins, these are heads of lists */
+  struct malloc_chunk* bk;
 
-  unsigned int         dirty;    /* unused for chunks, but declared so as */
-                                 /* to be conservative  wrt alignment */
-                                 /* For bins, the number of chunks placed */
-                                 /* on list since last consolidation */
 };
 
 typedef struct malloc_chunk* mchunkptr;
-typedef struct malloc_chunk* mbinptr;
+
+struct malloc_bin
+{
+  struct malloc_chunk hd;        /* dummy list header */
+  unsigned int        dirty;     /* True if maybe consolidatable */
+                                 /* Wasting a word here makes */
+                                 /* sizeof(bin) a power of 2, */
+                                 /* which makes size2bin() faster */
+};
+
+typedef struct malloc_bin* mbinptr;
 
 
-/*
-  sizes, alignment
-*/
+/*  sizes, alignments */
 
 
 #define SIZE_SZ                   (sizeof(unsigned int))
 #define MALLOC_MIN_OVERHEAD       (SIZE_SZ + SIZE_SZ)
 #define MALLOC_ALIGN_MASK         (MALLOC_MIN_OVERHEAD - 1)
 
-#define MINSIZE    (sizeof(struct malloc_chunk)) /* MUST == 16! */
+#define MINSIZE (sizeof(struct malloc_chunk) + SIZE_SZ) /* MUST == 16! */
 
 
 /* pad request bytes into a usable size */
@@ -277,139 +316,139 @@ static inline int aligned_OK(void* m)
 
 #define FIRSTBIN (&(av[0])) 
 
-static struct malloc_chunk  av[MAXBIN] = 
+static struct malloc_bin  av[MAXBIN] = 
 {
-  { 0, &(av[0]),  &(av[0]), 0 },
-  { 0, &(av[1]),  &(av[1]), 0 },
-  { 0, &(av[2]),  &(av[2]), 0 },
-  { 0, &(av[3]),  &(av[3]), 0 },
-  { 0, &(av[4]),  &(av[4]), 0 },
-  { 0, &(av[5]),  &(av[5]), 0 },
-  { 0, &(av[6]),  &(av[6]), 0 },
-  { 0, &(av[7]),  &(av[7]), 0 },
-  { 0, &(av[8]),  &(av[8]), 0 },
-  { 0, &(av[9]),  &(av[9]), 0 },
+  { { 0, &(av[0].hd),  &(av[0].hd) }, 0 },
+  { { 0, &(av[1].hd),  &(av[1].hd) }, 0 },
+  { { 0, &(av[2].hd),  &(av[2].hd) }, 0 },
+  { { 0, &(av[3].hd),  &(av[3].hd) }, 0 },
+  { { 0, &(av[4].hd),  &(av[4].hd) }, 0 },
+  { { 0, &(av[5].hd),  &(av[5].hd) }, 0 },
+  { { 0, &(av[6].hd),  &(av[6].hd) }, 0 },
+  { { 0, &(av[7].hd),  &(av[7].hd) }, 0 },
+  { { 0, &(av[8].hd),  &(av[8].hd) }, 0 },
+  { { 0, &(av[9].hd),  &(av[9].hd) }, 0 },
 
-  { 0, &(av[10]), &(av[10]), 0 },
-  { 0, &(av[11]), &(av[11]), 0 },
-  { 0, &(av[12]), &(av[12]), 0 },
-  { 0, &(av[13]), &(av[13]), 0 },
-  { 0, &(av[14]), &(av[14]), 0 },
-  { 0, &(av[15]), &(av[15]), 0 },
-  { 0, &(av[16]), &(av[16]), 0 },
-  { 0, &(av[17]), &(av[17]), 0 },
-  { 0, &(av[18]), &(av[18]), 0 },
-  { 0, &(av[19]), &(av[19]), 0 },
+  { { 0, &(av[10].hd), &(av[10].hd) }, 0 },
+  { { 0, &(av[11].hd), &(av[11].hd) }, 0 },
+  { { 0, &(av[12].hd), &(av[12].hd) }, 0 },
+  { { 0, &(av[13].hd), &(av[13].hd) }, 0 },
+  { { 0, &(av[14].hd), &(av[14].hd) }, 0 },
+  { { 0, &(av[15].hd), &(av[15].hd) }, 0 },
+  { { 0, &(av[16].hd), &(av[16].hd) }, 0 },
+  { { 0, &(av[17].hd), &(av[17].hd) }, 0 },
+  { { 0, &(av[18].hd), &(av[18].hd) }, 0 },
+  { { 0, &(av[19].hd), &(av[19].hd) }, 0 },
 
-  { 0, &(av[20]), &(av[20]), 0 },
-  { 0, &(av[21]), &(av[21]), 0 },
-  { 0, &(av[22]), &(av[22]), 0 },
-  { 0, &(av[23]), &(av[23]), 0 },
-  { 0, &(av[24]), &(av[24]), 0 },
-  { 0, &(av[25]), &(av[25]), 0 },
-  { 0, &(av[26]), &(av[26]), 0 },
-  { 0, &(av[27]), &(av[27]), 0 },
-  { 0, &(av[28]), &(av[28]), 0 },
-  { 0, &(av[29]), &(av[29]), 0 },
+  { { 0, &(av[20].hd), &(av[20].hd) }, 0 },
+  { { 0, &(av[21].hd), &(av[21].hd) }, 0 },
+  { { 0, &(av[22].hd), &(av[22].hd) }, 0 },
+  { { 0, &(av[23].hd), &(av[23].hd) }, 0 },
+  { { 0, &(av[24].hd), &(av[24].hd) }, 0 },
+  { { 0, &(av[25].hd), &(av[25].hd) }, 0 },
+  { { 0, &(av[26].hd), &(av[26].hd) }, 0 },
+  { { 0, &(av[27].hd), &(av[27].hd) }, 0 },
+  { { 0, &(av[28].hd), &(av[28].hd) }, 0 },
+  { { 0, &(av[29].hd), &(av[29].hd) }, 0 },
 
-  { 0, &(av[30]), &(av[30]), 0 },
-  { 0, &(av[31]), &(av[31]), 0 },
-  { 0, &(av[32]), &(av[32]), 0 },
-  { 0, &(av[33]), &(av[33]), 0 },
-  { 0, &(av[34]), &(av[34]), 0 },
-  { 0, &(av[35]), &(av[35]), 0 },
-  { 0, &(av[36]), &(av[36]), 0 },
-  { 0, &(av[37]), &(av[37]), 0 },
-  { 0, &(av[38]), &(av[38]), 0 },
-  { 0, &(av[39]), &(av[39]), 0 },
+  { { 0, &(av[30].hd), &(av[30].hd) }, 0 },
+  { { 0, &(av[31].hd), &(av[31].hd) }, 0 },
+  { { 0, &(av[32].hd), &(av[32].hd) }, 0 },
+  { { 0, &(av[33].hd), &(av[33].hd) }, 0 },
+  { { 0, &(av[34].hd), &(av[34].hd) }, 0 },
+  { { 0, &(av[35].hd), &(av[35].hd) }, 0 },
+  { { 0, &(av[36].hd), &(av[36].hd) }, 0 },
+  { { 0, &(av[37].hd), &(av[37].hd) }, 0 },
+  { { 0, &(av[38].hd), &(av[38].hd) }, 0 },
+  { { 0, &(av[39].hd), &(av[39].hd) }, 0 },
 
-  { 0, &(av[40]), &(av[40]), 0 },
-  { 0, &(av[41]), &(av[41]), 0 },
-  { 0, &(av[42]), &(av[42]), 0 },
-  { 0, &(av[43]), &(av[43]), 0 },
-  { 0, &(av[44]), &(av[44]), 0 },
-  { 0, &(av[45]), &(av[45]), 0 },
-  { 0, &(av[46]), &(av[46]), 0 },
-  { 0, &(av[47]), &(av[47]), 0 },
-  { 0, &(av[48]), &(av[48]), 0 },
-  { 0, &(av[49]), &(av[49]), 0 },
+  { { 0, &(av[40].hd), &(av[40].hd) }, 0 },
+  { { 0, &(av[41].hd), &(av[41].hd) }, 0 },
+  { { 0, &(av[42].hd), &(av[42].hd) }, 0 },
+  { { 0, &(av[43].hd), &(av[43].hd) }, 0 },
+  { { 0, &(av[44].hd), &(av[44].hd) }, 0 },
+  { { 0, &(av[45].hd), &(av[45].hd) }, 0 },
+  { { 0, &(av[46].hd), &(av[46].hd) }, 0 },
+  { { 0, &(av[47].hd), &(av[47].hd) }, 0 },
+  { { 0, &(av[48].hd), &(av[48].hd) }, 0 },
+  { { 0, &(av[49].hd), &(av[49].hd) }, 0 },
 
-  { 0, &(av[50]), &(av[50]), 0 },
-  { 0, &(av[51]), &(av[51]), 0 },
-  { 0, &(av[52]), &(av[52]), 0 },
-  { 0, &(av[53]), &(av[53]), 0 },
-  { 0, &(av[54]), &(av[54]), 0 },
-  { 0, &(av[55]), &(av[55]), 0 },
-  { 0, &(av[56]), &(av[56]), 0 },
-  { 0, &(av[57]), &(av[57]), 0 },
-  { 0, &(av[58]), &(av[58]), 0 },
-  { 0, &(av[59]), &(av[59]), 0 },
+  { { 0, &(av[50].hd), &(av[50].hd) }, 0 },
+  { { 0, &(av[51].hd), &(av[51].hd) }, 0 },
+  { { 0, &(av[52].hd), &(av[52].hd) }, 0 },
+  { { 0, &(av[53].hd), &(av[53].hd) }, 0 },
+  { { 0, &(av[54].hd), &(av[54].hd) }, 0 },
+  { { 0, &(av[55].hd), &(av[55].hd) }, 0 },
+  { { 0, &(av[56].hd), &(av[56].hd) }, 0 },
+  { { 0, &(av[57].hd), &(av[57].hd) }, 0 },
+  { { 0, &(av[58].hd), &(av[58].hd) }, 0 },
+  { { 0, &(av[59].hd), &(av[59].hd) }, 0 },
 
-  { 0, &(av[60]), &(av[60]), 0 },
-  { 0, &(av[61]), &(av[61]), 0 },
-  { 0, &(av[62]), &(av[62]), 0 },
-  { 0, &(av[63]), &(av[63]), 0 },
-  { 0, &(av[64]), &(av[64]), 0 },
-  { 0, &(av[65]), &(av[65]), 0 },
-  { 0, &(av[66]), &(av[66]), 0 },
-  { 0, &(av[67]), &(av[67]), 0 },
-  { 0, &(av[68]), &(av[68]), 0 },
-  { 0, &(av[69]), &(av[69]), 0 },
+  { { 0, &(av[60].hd), &(av[60].hd) }, 0 },
+  { { 0, &(av[61].hd), &(av[61].hd) }, 0 },
+  { { 0, &(av[62].hd), &(av[62].hd) }, 0 },
+  { { 0, &(av[63].hd), &(av[63].hd) }, 0 },
+  { { 0, &(av[64].hd), &(av[64].hd) }, 0 },
+  { { 0, &(av[65].hd), &(av[65].hd) }, 0 },
+  { { 0, &(av[66].hd), &(av[66].hd) }, 0 },
+  { { 0, &(av[67].hd), &(av[67].hd) }, 0 },
+  { { 0, &(av[68].hd), &(av[68].hd) }, 0 },
+  { { 0, &(av[69].hd), &(av[69].hd) }, 0 },
 
-  { 0, &(av[70]), &(av[70]), 0 },
-  { 0, &(av[71]), &(av[71]), 0 },
-  { 0, &(av[72]), &(av[72]), 0 },
-  { 0, &(av[73]), &(av[73]), 0 },
-  { 0, &(av[74]), &(av[74]), 0 },
-  { 0, &(av[75]), &(av[75]), 0 },
-  { 0, &(av[76]), &(av[76]), 0 },
-  { 0, &(av[77]), &(av[77]), 0 },
-  { 0, &(av[78]), &(av[78]), 0 },
-  { 0, &(av[79]), &(av[79]), 0 },
+  { { 0, &(av[70].hd), &(av[70].hd) }, 0 },
+  { { 0, &(av[71].hd), &(av[71].hd) }, 0 },
+  { { 0, &(av[72].hd), &(av[72].hd) }, 0 },
+  { { 0, &(av[73].hd), &(av[73].hd) }, 0 },
+  { { 0, &(av[74].hd), &(av[74].hd) }, 0 },
+  { { 0, &(av[75].hd), &(av[75].hd) }, 0 },
+  { { 0, &(av[76].hd), &(av[76].hd) }, 0 },
+  { { 0, &(av[77].hd), &(av[77].hd) }, 0 },
+  { { 0, &(av[78].hd), &(av[78].hd) }, 0 },
+  { { 0, &(av[79].hd), &(av[79].hd) }, 0 },
 
-  { 0, &(av[80]), &(av[80]), 0 },
-  { 0, &(av[81]), &(av[81]), 0 },
-  { 0, &(av[82]), &(av[82]), 0 },
-  { 0, &(av[83]), &(av[83]), 0 },
-  { 0, &(av[84]), &(av[84]), 0 },
-  { 0, &(av[85]), &(av[85]), 0 },
-  { 0, &(av[86]), &(av[86]), 0 },
-  { 0, &(av[87]), &(av[87]), 0 },
-  { 0, &(av[88]), &(av[88]), 0 },
-  { 0, &(av[89]), &(av[89]), 0 },
+  { { 0, &(av[80].hd), &(av[80].hd) }, 0 },
+  { { 0, &(av[81].hd), &(av[81].hd) }, 0 },
+  { { 0, &(av[82].hd), &(av[82].hd) }, 0 },
+  { { 0, &(av[83].hd), &(av[83].hd) }, 0 },
+  { { 0, &(av[84].hd), &(av[84].hd) }, 0 },
+  { { 0, &(av[85].hd), &(av[85].hd) }, 0 },
+  { { 0, &(av[86].hd), &(av[86].hd) }, 0 },
+  { { 0, &(av[87].hd), &(av[87].hd) }, 0 },
+  { { 0, &(av[88].hd), &(av[88].hd) }, 0 },
+  { { 0, &(av[89].hd), &(av[89].hd) }, 0 },
 
-  { 0, &(av[90]), &(av[90]), 0 },
-  { 0, &(av[91]), &(av[91]), 0 },
-  { 0, &(av[92]), &(av[92]), 0 },
-  { 0, &(av[93]), &(av[93]), 0 },
-  { 0, &(av[94]), &(av[94]), 0 },
-  { 0, &(av[95]), &(av[95]), 0 },
-  { 0, &(av[96]), &(av[96]), 0 },
-  { 0, &(av[97]), &(av[97]), 0 },
-  { 0, &(av[98]), &(av[98]), 0 },
-  { 0, &(av[99]), &(av[99]), 0 },
+  { { 0, &(av[90].hd), &(av[90].hd) }, 0 },
+  { { 0, &(av[91].hd), &(av[91].hd) }, 0 },
+  { { 0, &(av[92].hd), &(av[92].hd) }, 0 },
+  { { 0, &(av[93].hd), &(av[93].hd) }, 0 },
+  { { 0, &(av[94].hd), &(av[94].hd) }, 0 },
+  { { 0, &(av[95].hd), &(av[95].hd) }, 0 },
+  { { 0, &(av[96].hd), &(av[96].hd) }, 0 },
+  { { 0, &(av[97].hd), &(av[97].hd) }, 0 },
+  { { 0, &(av[98].hd), &(av[98].hd) }, 0 },
+  { { 0, &(av[99].hd), &(av[99].hd) }, 0 },
 
-  { 0, &(av[100]), &(av[100]), 0 },
-  { 0, &(av[101]), &(av[101]), 0 },
-  { 0, &(av[102]), &(av[102]), 0 },
-  { 0, &(av[103]), &(av[103]), 0 },
-  { 0, &(av[104]), &(av[104]), 0 },
-  { 0, &(av[105]), &(av[105]), 0 },
-  { 0, &(av[106]), &(av[106]), 0 },
-  { 0, &(av[107]), &(av[107]), 0 },
-  { 0, &(av[108]), &(av[108]), 0 },
-  { 0, &(av[109]), &(av[109]), 0 },
+  { { 0, &(av[100].hd), &(av[100].hd) }, 0 },
+  { { 0, &(av[101].hd), &(av[101].hd) }, 0 },
+  { { 0, &(av[102].hd), &(av[102].hd) }, 0 },
+  { { 0, &(av[103].hd), &(av[103].hd) }, 0 },
+  { { 0, &(av[104].hd), &(av[104].hd) }, 0 },
+  { { 0, &(av[105].hd), &(av[105].hd) }, 0 },
+  { { 0, &(av[106].hd), &(av[106].hd) }, 0 },
+  { { 0, &(av[107].hd), &(av[107].hd) }, 0 },
+  { { 0, &(av[108].hd), &(av[108].hd) }, 0 },
+  { { 0, &(av[109].hd), &(av[109].hd) }, 0 },
 
-  { 0, &(av[110]), &(av[110]), 0 },
-  { 0, &(av[111]), &(av[111]), 0 },
-  { 0, &(av[112]), &(av[112]), 0 },
-  { 0, &(av[113]), &(av[113]), 0 },
-  { 0, &(av[114]), &(av[114]), 0 },
-  { 0, &(av[115]), &(av[115]), 0 },
-  { 0, &(av[116]), &(av[116]), 0 },
-  { 0, &(av[117]), &(av[117]), 0 },
-  { 0, &(av[118]), &(av[118]), 0 },
-  { 0, &(av[119]), &(av[119]), 0 }
+  { { 0, &(av[110].hd), &(av[110].hd) }, 0 },
+  { { 0, &(av[111].hd), &(av[111].hd) }, 0 },
+  { { 0, &(av[112].hd), &(av[112].hd) }, 0 },
+  { { 0, &(av[113].hd), &(av[113].hd) }, 0 },
+  { { 0, &(av[114].hd), &(av[114].hd) }, 0 },
+  { { 0, &(av[115].hd), &(av[115].hd) }, 0 },
+  { { 0, &(av[116].hd), &(av[116].hd) }, 0 },
+  { { 0, &(av[117].hd), &(av[117].hd) }, 0 },
+  { { 0, &(av[118].hd), &(av[118].hd) }, 0 },
+  { { 0, &(av[119].hd), &(av[119].hd) }, 0 }
 };
 
 /*
@@ -428,6 +467,8 @@ static inline mbinptr size2bin(unsigned int sz)
 
 /* counts maintained if MALLOC_STATS defined */
 
+#ifdef MALLOC_STATS
+
 static unsigned int sbrked_mem;
 static unsigned int requested_mem;
 static unsigned int malloced_mem;
@@ -449,9 +490,6 @@ static unsigned int n_split;
 static unsigned int n_consol;
 
 
-/* stat maintainers, for convenience */
-
-#ifdef MALLOC_STATS
 static void do_malloc_stats(const mchunkptr p)
 {
   ++n_mallocs;
@@ -471,62 +509,32 @@ static void do_free_stats(const mchunkptr p)
 #endif
 
 
+
+/* maintaining INUSE via size field */
+
+
+#define inuse(p)       ((p)->size & INUSE)
+#define set_inuse(p)   ((p)->size |= INUSE)
+#define clear_inuse(b) ((p)->size &= ~INUSE)
+
   
 /* operations on  malloc_chunk addresses */
 
 
 /* return ptr to next physical malloc_chunk */
 
-static inline mchunkptr next_chunk(const mchunkptr p)
-{
-  return (mchunkptr)((char*)(p) + p->size);
-}
-
+#define next_chunk(p) ((mchunkptr)((char*)(p) + (p)->size))
 
 /* return ptr to previous physical malloc_chunk */
 
-static inline mchunkptr prev_chunk(const mchunkptr p)
-{
-  return (mchunkptr)((char*)p - ((((int*)(p))[-1]) & ~(INUSE)));
-}
+#define prev_chunk(p) ((mchunkptr)((char*)(p)-((((int*)(p))[-1]) & ~(INUSE))))
 
 /* place size at front and back of chunk */
 
+
 static inline void set_size(mchunkptr p, unsigned int sz)
 {
-  p->size = *((int*)((char*)p + sz - SIZE_SZ)) = sz;
-}
-
-
-/* return pointer to a new malloc_chunk being split off */
-
-static inline mchunkptr split(mchunkptr p, unsigned int offset)
-{
-  unsigned int room = p->size - offset;
-  mchunkptr t = (mchunkptr)((char*)(p) + offset);
-  t->size = *((int*)((char*)t + room    - SIZE_SZ)) = room;
-  p->size = *((int*)((char*)p + offset  - SIZE_SZ)) = offset;
-  UPDATE_STATS (++n_split);
-  return t;
-}
-
-
-
-/* maintaining INUSE via size field */
-
-static inline unsigned int inuse(const mchunkptr p)
-{
-  return p->size & INUSE;
-}
-
-static inline void set_inuse(mchunkptr p)
-{
-  p->size |= INUSE;
-}
-
-static inline void clear_inuse(mchunkptr p)
-{
-  p->size &= ~INUSE;
+  p->size = *((int*)((char*)(p) + sz - SIZE_SZ)) = sz;
 }
 
 
@@ -536,17 +544,17 @@ static inline void clear_inuse(mchunkptr p)
 
 static inline void* chunk2mem(mchunkptr p) 
 { 
-  p->size |= INUSE;
-  return (void*)((char*)p + SIZE_SZ); 
+  set_inuse(p);
+  return (void*)((char*)(p) + SIZE_SZ); 
 }
 
 static inline mchunkptr mem2chunk(void* mem) 
 { 
-  mchunkptr p = (mchunkptr)((char*)mem - SIZE_SZ); 
+  mchunkptr p = (mchunkptr)((char*)(mem) - SIZE_SZ); 
 
   /* a quick sanity check */
   unsigned int sz = p->size & ~(INUSE);
-  if (p->size == sz || sz != *((int*)((char*)p + sz - SIZE_SZ)))
+  if (p->size == sz || sz != *((int*)((char*)(p) + sz - SIZE_SZ)))
     malloc_user_error();
 
   p->size = sz;   /* clears INUSE */
@@ -563,12 +571,11 @@ static inline mchunkptr mem2chunk(void* mem)
 static mbinptr malloc_maxbin = FIRSTBIN;
 
 
-/* The designated victim. */
+/* The designated victim bin. */
+
+static mbinptr malloc_victim = FIRSTBIN;
 
-/* The victim is always checked for size, so it is OK to use bin heads */
-/* (which always have size == 0) to represent unusable victims */
 
-static mchunkptr malloc_victim = FIRSTBIN;
 
 
 /* operations on lists inside bins */
@@ -587,52 +594,81 @@ static inline void unlink(mchunkptr p)
 }
 
 
+
+/* split a chunk and place on the back of a list */
+
+static inline void split(mchunkptr p, unsigned int offset)
+{
+  unsigned int room = p->size - offset;
+  if (room >= MINSIZE)
+  {
+    mbinptr   bn = size2bin(room);                  /* new bin */
+    mchunkptr h  = &(bn->hd);                       /* its head */
+    mchunkptr b  = h->bk;                           /* old back element */
+    mchunkptr t = (mchunkptr)((char*)(p) + offset); /* remaindered chunk */
+    
+    /* set size */
+    t->size = *((int*)((char*)(t) + room    - SIZE_SZ)) = room;
+
+    /* link up */
+    t->bk = b;  t->fd = h;  h->bk = b->fd = t;
+    
+    /* adjust maxbin (h == b means was empty) */
+    if (h == b && bn > malloc_maxbin) malloc_maxbin = bn; 
+
+    /* use as next victim */
+    malloc_victim = bn;
+    
+    /* adjust size of chunk to be returned */
+    p->size = *((int*)((char*)(p) + offset  - SIZE_SZ)) = offset;
+
+    UPDATE_STATS ((++n_split, ++n_avail));
+  }
+}
+
+
+
+/* place a consolidated chunk on the back of a list */
+/* like above, except no split */
+
+static inline void consollink(mchunkptr p)
+{
+  mbinptr   bn = size2bin(p->size);
+  mchunkptr h  = &(bn->hd);
+  mchunkptr b  = h->bk;
+
+  p->bk = b;  p->fd = h;  h->bk = b->fd = p;
+
+  if (h == b && bn > malloc_maxbin) malloc_maxbin = bn; 
+
+  malloc_victim = bn;        /* use as next victim */
+
+  UPDATE_STATS(++n_avail);
+}
+
+
 /* place a freed chunk on the front of a list */
 
 static inline void frontlink(mchunkptr p)
 {
   mbinptr   bn = size2bin(p->size);
-  mchunkptr h  = (mchunkptr)(bn);
+  mchunkptr h  = &(bn->hd);
   mchunkptr f  = h->fd;
 
   p->bk = h;  p->fd = f;  f->bk = h->fd = p;
 
-  bn->dirty++;   /* mark that bin can be consolidated */
+  if (h == f && bn > malloc_maxbin) malloc_maxbin = bn;  
 
-  if (h == f && bn > malloc_maxbin)   /* h==f means was empty */
-    malloc_maxbin = bn;  
+  bn->dirty = 1;
 
   UPDATE_STATS(++n_avail);
 }
 
 
-/* place a split or consolidated chunk on the back of a list */
-
-static inline void backlink(mchunkptr p)
-{
-  mbinptr   bn = size2bin(p->size);
-  mchunkptr h  = (mchunkptr)(bn);
-  mchunkptr b  = h->bk;
-
-  p->bk = b;  p->fd = h;  h->bk = b->fd = p;
-
-  malloc_victim = p;        /* use as next victim */
-  bn->dirty++;              /* mark for consolidation */
-
-  if (h == b && bn > malloc_maxbin) malloc_maxbin = bn; 
-
-  UPDATE_STATS(++n_avail);
-}
-
 
 /* Dealing with sbrk */
 
 
-/* A good size to call sbrk with, == page size, as computed in morecore */
-
-static unsigned int best_sbrk_size;
-
-
 /* To link consecutive sbrk regions when possible */
 
 static int* last_sbrk_end;
@@ -640,9 +676,14 @@ static int* last_sbrk_end;
 
 /*  who to call when sbrk returns failure */
 
+#ifndef NO_NEW_HANDLER
 typedef volatile void (*vfp)();
+#ifdef __cplusplus
+extern "C" vfp __new_handler;
+#else
 extern vfp __new_handler;
-
+#endif
+#endif
 
 static mchunkptr malloc_from_sys(unsigned nb)
 {
@@ -650,40 +691,18 @@ static mchunkptr malloc_from_sys(unsigned nb)
   unsigned int sbrk_size;
   int* ip;
   
-  if (best_sbrk_size == 0)  /* first time through */
-  {
-    best_sbrk_size = getpagesize();
-    if (best_sbrk_size < 4096)  /* use something reasonable! */
-      best_sbrk_size = 4096;
-  }
-
   /* Minimally, we need to pad with enough space */
   /* to place dummy size/use fields to ends if needed */
-  /* and a remander chunk of at least MINSIZE */
-  /* (malloc expects that there will be a remainder) */
-  /* then round to page size. */
 
-  /* But there is little harm in being a little wasteful here */
-  /* and grabbing one more page than minimally needed in anticipation */
-  /* of more requests, since the other routines handle large chunks */
-  /* of space better than small ones. (They will need to scan */
-  /* fewer chunks to consolidate enough to fulfill requests, etc.) */
-
-#ifdef MIN_MALLOC_SBRK_WASTE
-
-  sbrk_size = ((nb + best_sbrk_size - 1 + SIZE_SZ + SIZE_SZ + MINSIZE) 
-               / best_sbrk_size) * best_sbrk_size;
-
-#else
-
-  sbrk_size = (1 + (nb + best_sbrk_size-1) / best_sbrk_size) * best_sbrk_size;
-
-#endif
+  sbrk_size = ((nb + SBRK_UNIT - 1 + SIZE_SZ + SIZE_SZ) 
+               / SBRK_UNIT) * SBRK_UNIT;
 
   ip = (int*)(sbrk(sbrk_size));
   if ((char*)ip == (char*)(-1)) /* sbrk returns -1 on failure */
   {
+#ifndef NO_NEW_HANDLER
     (*__new_handler) ();
+#endif
     return 0;
   }
 
@@ -703,7 +722,6 @@ static mchunkptr malloc_from_sys(unsigned nb)
     /* Note we can get away with only 1 word, not MINSIZE overhead here */
 
     *ip++ = SIZE_SZ | INUSE;
-
     
     p = (mchunkptr)ip;
     set_size(p,sbrk_size - (SIZE_SZ + SIZE_SZ)); 
@@ -714,7 +732,7 @@ static mchunkptr malloc_from_sys(unsigned nb)
     mchunkptr l;  
 
     /* We can safely make the header start at end of prev sbrked chunk. */
-    /* We will still have 1 word left at the end from a previous call */
+    /* We will still have space left at the end from a previous call */
     /* to place the end marker, below */
 
     p = (mchunkptr)(last_sbrk_end);
@@ -738,114 +756,93 @@ static mchunkptr malloc_from_sys(unsigned nb)
   last_sbrk_end = (int*)((char*)p + p->size);
   *last_sbrk_end = SIZE_SZ | INUSE;
 
-
-  /* make p look like it's in a list, so unlink will work. */
-  /* (malloc will properly relink remainder) */
-
-  p->fd = p->bk = p;
-
   UPDATE_STATS((++n_avail, ++n_malloc_chunks));
-
+  
   return p;
 }
 
 
 
-/* Consolidate all chunks in bin b. */
-/* If this creates a chunk big enough to satisfy */
-/* current malloc request, return it */
+/* Consolidate dirty bins. */
+/* Stop if found a chunk big enough to satisfy current malloc request */
 
 /* (It requires much less bookkeeping to consolidate entire bins */
 /* at once than to keep records of which chunks might be */
 /* consolidatable. So long as the lists are short, which we */
 /* try to ensure via small bin ranges, there is little wasted effort.) */
 
-static mchunkptr malloc_consolidate_bin(mbinptr b, unsigned int nb)
-{
-  mchunkptr h = (mchunkptr)(b);   /* head of list */
-  mchunkptr p = h->bk;            /* chunk traverser */
-  mchunkptr found = 0;            /* to hold usable consolidated chunk */
-  
-  b->dirty = 0;                   /* mark that bin is clean */
-  
-  /* traverse each chunk, looking for consolidations */
-  /* proceed backwards to avoid rescanning relinked chunks */
-
-  while (p != h)
-  {
-    mchunkptr nextp = p->bk;        /* save, in case of relinks */
-
-    mchunkptr l = prev_chunk(p);
-    mchunkptr n = next_chunk(p);
-
-    if (!inuse(l) || !inuse(n))
-    {
-      unlink(p);
-
-      while (!inuse(l)) /* consolidate backward */
-      {
-        if (l == nextp) nextp = l->bk;
-        unlink(l);
-        set_size(l, l->size + p->size);
-        p = l;
-        l = prev_chunk(p);
-        UPDATE_STATS (++n_consol);
-      }
-      
-      while (!inuse(n)) /* and forward */
-      {
-        if (n == nextp) nextp = n->bk;
-        unlink(n);
-        set_size(p, p->size + n->size);
-        n = next_chunk(p);
-        UPDATE_STATS (++n_consol);
-      }
-      
-      if (p->size >= nb) found = p;
-
-      backlink(p);
-    }
-
-    UPDATE_STATS(++n_malloc_chunks);
-    p = nextp;
-  }
-
-  return found;  
-}
-
-
-
-
-/* Consolidate dirty bins. */
-/* Stop if found a chunk big enough to satisfy current malloc request */
-/* If can't consolidate to fill request, get more from sys */
-
 static mchunkptr malloc_find_space(unsigned int nb)
 {
   mbinptr b;
+  mchunkptr usable = 0;
 
-  /* traverse backwards so as not to rescan chunks we */
-  /* just put in bigger bins in previous iterations  */
+  /* first, re-adjust max used bin */
+
+  while (malloc_maxbin >= FIRSTBIN && 
+         malloc_maxbin->hd.bk == &(malloc_maxbin->hd))
+  {
+    malloc_maxbin->dirty = 0;
+    --malloc_maxbin;
+  }
 
   for (b = malloc_maxbin; b >= FIRSTBIN; --b)
   {
     UPDATE_STATS(++n_malloc_bins);
-   
+
     if (b->dirty)
     {
-      if (b->bk == b)    /* empty - just clear dirty */
-        b->dirty = 0; 
-      else
+      mchunkptr h = &(b->hd);         /* head of list */
+      mchunkptr p = h->bk;            /* chunk traverser */
+
+      while (p != h)
       {
-        mchunkptr p = malloc_consolidate_bin(b, nb);
-        if (p != 0)
-          return p;
+        mchunkptr nextp = p->bk;       /* save, in case of relinks */
+        int consolidated = 0;          /* only unlink/relink if consolidated */
+
+        mchunkptr t;
+        while (!inuse(t = prev_chunk(p))) /* consolidate backward */
+        {
+          if (!consolidated) { consolidated = 1; unlink(p); }
+          if (t == nextp) nextp = t->bk;
+          unlink(t);
+          set_size(t, t->size + p->size);
+          p = t;
+          UPDATE_STATS (++n_consol);
+        }
+        
+        while (!inuse(t = next_chunk(p))) /* consolidate forward */
+        {
+          if (!consolidated) { consolidated = 1; unlink(p); }
+          if (t == nextp) nextp = t->bk;
+          unlink(t);
+          set_size(p, p->size + t->size);
+          UPDATE_STATS (++n_consol);
+        }
+
+        if (consolidated)
+        {
+          consollink(p);
+          if (p->size >= nb)
+            usable = p;
+        }
+
+        UPDATE_STATS(++n_malloc_chunks);
+        p = nextp;
+
       }
+
+      b->dirty = 0;
+
+      if (usable != 0) 
+      {
+        unlink(usable);
+        return usable;
+      }
+
     }
   }
 
-
-  /* nothing available -- get some more */
+  /* nothing available - sbrk some more */
 
   return malloc_from_sys(nb);
 }
@@ -858,83 +855,83 @@ void* malloc(unsigned int bytes)
 {
   unsigned int nb  = request2size(bytes);  /* padded request size */
   mbinptr      b   = size2bin(nb);         /* corresponding bin */
-  mchunkptr    hd  = (mchunkptr)(b);       /* head of its list */
-  mchunkptr    p   = hd->fd;               /* chunk traverser */
+  mchunkptr    hd  = &(b->hd);             /* head of its list */
+  mchunkptr    p;                          /* chunk traverser */
 
   UPDATE_STATS((requested_mem+=bytes, ++n_malloc_bins));
 
-  /* if bin non-empty, try a (near) exact match */
+  /* Try a (near) exact match in own bin */
 
-  if (p != hd) 
+  for (p = hd->fd; p != hd; p = p->fd)
   {
-    do
+    UPDATE_STATS(++n_malloc_chunks);
+
+    if (p->size >= nb)
     {
-      UPDATE_STATS(++n_malloc_chunks);
-
-      if (p->size >= nb)         /* found one - take without splitting */
-      {
-
-        if (p == malloc_victim)  /* reset victim for next time */
-          malloc_victim = (p->bk == hd)? p->fd : p->bk;
-
-        unlink(p);
-        UPDATE_STATS(do_malloc_stats(p));
-        return chunk2mem(p);
-      }
-
-    } while ((p = p->fd) != hd);
-
-    /* If no match, maybe consolidate in order to minimize future searching */
-
-    /* This is just a heuristic, and the criterion is just a magic */
-    /* number, but it seems robust across test cases */
-
-    if (b->dirty > 4) 
-    { 
-      if ((p = malloc_consolidate_bin(b, nb)) != 0)
-        malloc_victim = p; /* If we happened to find something, use it */
+      unlink(p);
+      goto found;
     }
-
   }
 
-
   /* No exact match, so select a victim */
 
-  p = malloc_victim; 
   UPDATE_STATS(++n_malloc_chunks);
 
-  if (p->size < nb)  
+  p = malloc_victim->hd.bk; 
+  if (p->size >= nb)
+    unlink(p);
+  else
   {
-    for (;;)                                   /* search bins for new victim */
+    /* Find a new victim. */
+
+    for (;;)
     {
       UPDATE_STATS(++n_malloc_bins);
-      if (malloc_maxbin <= b)                  /* nowhere to look */
+      ++b;
+      p = b->hd.bk;
+      if (p != &(b->hd)) 
       {
-        if ((p = malloc_find_space(nb)) == 0)  /* consolidate or sbrk */
-          return 0;                            /* out of space! */
+        /* p must be usable since it's in bigger bin */
+        /* Consolidate, so it can serve as a victim longer */
+
+        mchunkptr t;
+
+        unlink(p);
+
+        while (!inuse(t = prev_chunk(p))) /* consolidate backward */
+        {
+          unlink(t);
+          set_size(t, t->size + p->size);
+          p = t;
+          UPDATE_STATS (++n_consol);
+        }
+        
+        while (!inuse(t = next_chunk(p))) /* consolidate forward */
+        {
+          unlink(t);
+          set_size(p, p->size + t->size);
+          UPDATE_STATS (++n_consol);
+        }
+
+        malloc_victim = b;  /* record this bin for next time */
+        break;
+      }
+      else if (b > malloc_maxbin)
+      {
+        p = malloc_find_space(nb);
+        if (p == 0)
+          return 0;
         else
           break;
       }
-      else if ((p = malloc_maxbin->bk) != malloc_maxbin) /* found one  */
-      {
-        break;        /* (p MUST be usable since it's in a bigger bin) */
-      }
-      else
-      {
-        --malloc_maxbin;              /* maxbin must be empty, adjust */
-      }
     }
   }
 
   /* use what we found */
 
-  unlink(p);
-
-  if (p->size - nb >= MINSIZE)
-    backlink(split(p, nb));        /* (backlink resets victim) */
-  else                           
-    malloc_victim = p->bk;         /* make sure victim OK for next time */
+ found:
 
+  split(p, nb); 
   UPDATE_STATS(do_malloc_stats(p));
   return chunk2mem(p);
 }
@@ -970,8 +967,17 @@ void cfree(void *mem, unsigned int n, unsigned int elem_size)
 
 unsigned int malloc_usable_size(void* mem)
 {
-  mchunkptr p = (mchunkptr)((char*)mem - SIZE_SZ); 
-  return (p->size & ~INUSE) - MALLOC_MIN_OVERHEAD;
+  if (mem == 0)
+    return 0;
+  else
+  {
+    mchunkptr p = (mchunkptr)((char*)(mem) - SIZE_SZ); 
+    unsigned int sz = p->size & ~(INUSE);
+    if (p->size == sz || sz != *((int*)((char*)(p) + sz - SIZE_SZ)))
+      return 0;
+    else
+      return sz - MALLOC_MIN_OVERHEAD;
+  }
 }
 
 
@@ -985,42 +991,37 @@ void* realloc(void* mem, unsigned int bytes)
     unsigned int nb      = request2size(bytes);
     mchunkptr    p       = mem2chunk(mem);
     unsigned int oldsize = p->size;
-    int          room    = oldsize - nb;
+    int          room;
+    mchunkptr    nxt;
 
-    UPDATE_STATS(++n_reallocs);
+    UPDATE_STATS((++n_reallocs, requested_mem += bytes-oldsize));
     
-    while (room < 0) /* try to expand */
+    /* try to expand (even if already big enough), to clean up chunk */
+
+    while (!inuse(nxt = next_chunk(p)))
     {
-      mchunkptr nxt = next_chunk(p);
-      if (inuse(nxt))
-        break;
-      else
-      {
-        UPDATE_STATS ((malloced_mem += nxt->size, ++n_consol));
-        if (nxt == malloc_victim) /* maintain usable victim */
-          malloc_victim = nxt->bk;
-        unlink(nxt);
-        set_size(p, p->size + nxt->size);
-        room = p->size - nb;
-      }
+      UPDATE_STATS ((malloced_mem += nxt->size, ++n_consol));
+      unlink(nxt);
+      set_size(p, p->size + nxt->size);
     }
 
+    room = p->size - nb;
     if (room >= 0)
     {
-      if (room >= MINSIZE) 
+      if (room >= MINSIZE)
       {
-        backlink(split(p, nb));
-        UPDATE_STATS(freed_mem += room);
+        split(p, nb);
+        UPDATE_STATS(malloced_mem -= room);
       }
+      
       return chunk2mem(p);
     }
     else /* do the obvious */
     {
       void* newmem;
-      unsigned int s = (oldsize <= nb)? oldsize : nb;
       set_inuse(p);    /* don't let malloc consolidate us yet! */
       newmem = malloc(nb);
-      bcopy(mem, newmem, s - SIZE_SZ);
+      bcopy(mem, newmem, oldsize - SIZE_SZ);
       free(mem);
       UPDATE_STATS(++n_reallocs_with_copy);
       return newmem;
@@ -1032,6 +1033,9 @@ void* realloc(void* mem, unsigned int bytes)
 
 void malloc_stats()
 {
+#ifndef MALLOC_STATS
+}
+#else
   int i;
   mchunkptr p;
   double nm = (double)(n_mallocs + n_reallocs);
@@ -1088,12 +1092,12 @@ void malloc_stats()
   fprintf(stderr, "\nfree chunks:\n");
   for (i = 0; i < MAXBIN; ++i)
   {
-    p = av[i].fd;
-    if (p != &av[i])
+    p = av[i].hd.fd;
+    if (p != &(av[i].hd))
     {
       unsigned int count = 1;
       unsigned int sz = p->size;
-      for (p = p->fd; p != &av[i]; p = p->fd)
+      for (p = p->fd; p != &(av[i].hd); p = p->fd)
       {
         if (p->size == sz)
           ++count;
@@ -1110,5 +1114,6 @@ void malloc_stats()
     }
   }
 }
+#endif /* MALLOC_STATS */
  
-#endif
+#endif /* NO_LIBGXX_MALLOC */
